#ifdef H8_3067
  #include <h8/reg3067.h>
#endif
#ifdef H8_3048
  #include <h8/reg3048.h>
#endif
#include	"ata.h"
#include	"fs.h"
static volatile unsigned short	*atabase0, *atabase1;
static unsigned char	hdddrive;
static unsigned int	hddsize[2];

#ifdef WITH_FS
  extern diskinfo	hd0, hd1, ee0;
  int			get_diskinfo(int, diskinfo*);
#endif

int device_selection() {
	int	i, cnt;

	cnt = 0;
	while(ATA_SR & SR_BSY) {
		if(cnt++ > 200000) return -1;
	}
	ATA_DHR = hdddrive;
	for(i = 0;i < 2;i++);
	cnt = 0;
	while((ATA_SR &(SR_BSY | SR_DRDY)) != SR_DRDY){
		if(cnt++ > 200000) return -1;
	}
	return 0;
}

int non_data_command(unsigned short command, int count, int lba) {
	if(device_selection() == -1) return -1;
	ATA_CR = command;
	ATA_SCR = count;
	ATA_SNR = lba & 0xff;
	ATA_CYLL = (lba >> 8) & 0xff;
	ATA_CYLH = (lba >> 16) & 0xff;
	ATA_DHR |= DHR_LBA | ((lba >> 24) & 0xf);
	while(ATA_SR & SR_BSY);
	if(ATA_SR & (SR_DF | SR_ERR)) return -1;
	return 0;
}

int ata_sel_drive(int drive) {
	hdddrive = (drive == 0) ? 0 : DHR_DEV;
	if(drive == 0)
		non_data_command(SetMultiple, 128, 0);
	else
		non_data_command(SetMultiple, 128, 0);
}

int ata_identify(short *data) {
	int i, s;

	if(device_selection() == -1) return -1;
	ATA_CR = IdentifyDevice;
	while(ATA_ASR & SR_BSY);
	if(ATA_SR & (SR_DF | SR_ERR)) return -1;
	s = ATA_SR;
	if(s & SR_DREQ) {
		for(i = 0;i < 256;i++) data[i] =  ATA_DR;
	}
	return 0;
}

int data_in_command(unsigned short command, char *data, int lba, int count) {
	int	i, s;
	short 	value;

	if(device_selection() == -1) return -1;
	ATA_SCR = count;
	ATA_SNR = lba & 0xff;
	ATA_CYLL = (lba >> 8) & 0xff;
	ATA_CYLH = (lba >> 16) & 0xff;
	ATA_DHR |= DHR_LBA | ((lba >> 24) & 0xf);
	ATA_CR = command;
	while(ATA_ASR & SR_BSY);
	if(ATA_SR & (SR_DF | SR_ERR)) return -1;
	s = ATA_SR;
	if(s & SR_DREQ) {
		for(i = 0;i < 512 * count;) {
			value =  ATA_DR;
			data[i++] = value & 0xff;
			data[i++] = (value >> 8) & 0xff;
		}
		return 0;
	}
	return -1;
}

int data_out_command(unsigned short command, char *data, int lba, int count) {
	int	i, s;
	short 	value;

	if(device_selection() == -1) return -1;
	ATA_CR = command;
	ATA_SCR = count;
	ATA_SNR = lba & 0xff;
	ATA_CYLL = (lba >> 8) & 0xff;
	ATA_CYLH = (lba >> 16) & 0xff;
	ATA_DHR |= DHR_LBA | ((lba >> 24) & 0xf);
	while(ATA_ASR & SR_BSY);
	if(ATA_SR & (SR_DF | SR_ERR)) return -1;
	s = ATA_SR;
	if(s & SR_DREQ) {
		for(i = 0;i < 512 * count;) {
			value = data[i++] & 0xff;
			value |= (data[i++] << 8) & 0xff00;
			ATA_DR = value;
		}
		while(ATA_ASR & SR_BSY);
		if(ATA_SR & (SR_DF | SR_ERR)) return -1;
		return 0;
	}
	return -1;
}

int ata_driveinfo(char *name, int *chs) {
	unsigned short	pio;
	unsigned short	data;
	int		i, s, cnt, lba;
	unsigned short	*hddname;

	if(device_selection() == -1) return -1;
	hddname = (unsigned short*)name;
	ATA_CR = IdentifyDevice;
	cnt = 0;
	while(ATA_ASR & SR_BSY) {
		if(cnt++ > 200000) return -1;
	}
	if(ATA_SR & (SR_DF | SR_ERR)) return -1;
	s = ATA_SR;
	if(s & SR_DREQ) {
		for(i = 0;i < 256;i++) {
			data = ATA_DR;
			if(i >= 27 && i < 47) hddname[i - 27] = data;
			switch(i) {
			case 54:
				chs[1] = data;
				break;
			case 55:
				chs[0] = data;
				break;
			case 56:
				chs[2] = data;
				break;
			case 64:
				pio = data & 0xff;
				break;
			}
		}
		hddsize[(hdddrive == 0) ? 0 : 1] = chs[0] * chs[1] * chs[2];
		return pio;
	}
	return -1;
}

int get_hddsize() {
	if(hdddrive == 0)
		return hddsize[0];
	else
		return hddsize[1];
}

int ata_reset() {
	int i, cnt;

	cnt = 0;
	ATA_DCR = DCR_SRST;
	for(i = 0;i < 7;i++);
	ATA_DCR = 0;
	for(i = 0;i < 3000;i++);
	while(ATA_SR & SR_BSY) {
		if(cnt++ > 200000) return -1;
	}
	return 0;
}

int ata_setup(int base0, int base1) {
	char	name[40], bw, msk;
	int	i, ret, chs[3], adr, inc;

#ifdef H8_300H
	inc = adr = 0x200000;
	bw = 0;
	msk = 2;
	for(i = 1;i < 8;i++) {
		if(base0 >= adr && base0 < (adr + inc)) bw |= msk;
		if(base1 >= adr && base1 < (adr + inc)) bw |= msk;
		msk <<= 1;
		adr += inc;
	}
	ABWCR &= ~bw;
#endif
	atabase0 = (volatile unsigned short*)base0;
	atabase1 = (volatile unsigned short*)base1;
	hdddrive = 0;
	if(ata_reset() == -1) return 0;
	ATA_DCR = DCR_nIEN;
	hddsize[0] = 0;
	hddsize[1] = 0;
	ret = 0;
	ata_sel_drive(0);
	if(ata_driveinfo(name, chs) >= 0) ret |= 1;
	ata_sel_drive(1);
	if(ata_driveinfo(name, chs) >= 0) ret |= 2;
	ata_sel_drive(0);
#ifdef WITH_FS
	if(ret & 1) get_diskinfo(HD0, &hd0);
	if(ret & 2) get_diskinfo(HD1, &hd1);
#endif
	return ret;
}
