#include <ctype.h>
#define BLKSIZ	256
#define NAMESIZ	12
#if defined(H8_3069) || defined(H8_3068) || defined(H8_3052) || defined(SH2_7045)
  #define INTER_ROM
#endif
#ifdef INTER_ROM
  #if defined(H8_3069) || defined(H8_3052)
    #define UNITMAX	7
  #endif
  #ifdef H8_3068
    #define UNITMAX	5
  #endif
  #ifdef SH2_7045
    #define UNITMAX	7
    #define UNITSIZ	0x8000
    #define UNITMIN	2
  #endif
  #if defined(H8_3069) || defined(H8_3068) || defined(H8_3052)
    #define UNITSIZ	0x10000
    #define UNITMIN	1
  #endif
  #define rom_write	write_flush
  #define rom_read	memcpy
  #define DATATYPE	char*
#else
  extern int	eepsize;
  #define UNITSIZ	0x4000
  #define UNITMIN	0
  #define UNITMAX	(eepsize / UNITSIZ - 1)
  #define rom_read	eep_read
  #define rom_write	eep_write
  #define DATATYPE	int
#endif
typedef struct {
	int	size;
	unsigned char	name[NAMESIZ];
} entry;
void	deflag();

int erase_unit(int unit) {
	char	buf[sizeof(entry)];
	int	ret;

#ifdef INTER_ROM
	ret = erase_flush((char*)(unit * UNITSIZ));
#else
	int	i;

	for(i = 0;i < sizeof(entry);i++) buf[i] = 0xff;
	for(i = 0;i < UNITSIZ / BLKSIZ;i++) {
		rom_write(i * sizeof(entry) + unit * UNITSIZ, buf, sizeof(entry));
	}
	ret = 0;
#endif
	strcpy(buf, "  easyfs");
	rom_write((DATATYPE)(unit * UNITSIZ), buf, 8);
	return ret;
}

int check_fd(int fd) {
	entry	ent;
	int	unit, unitfd, minfd, maxfd;

	minfd = (UNITSIZ * sizeof(entry)) / (BLKSIZ * BLKSIZ);
	maxfd = UNITSIZ / BLKSIZ;
	unit = fd / maxfd;
	unitfd = fd % maxfd;
	if(unit < UNITMIN || unit > UNITMAX) return -1;
	if(unitfd < minfd || unitfd >= maxfd) return -1;
	rom_read((char*)(&ent), (DATATYPE)(unitfd * sizeof(entry) + unit * UNITSIZ), sizeof(entry));
	if(ent.name[0] == 0 || ent.name[0] == 0xff) return -1;
	return 0;
}

int easy_format(char c) {
	int	unit;

	for(unit = UNITMIN;unit <= UNITMAX;unit++) {
		if(erase_unit(unit) == -1) return -1;
	}
	return 0;
}

int easy_max() {
	char	buf[8];

	rom_read(buf, (DATATYPE)(UNITMIN * UNITSIZ), 8);
	if(memcmp(&buf[2], "easyfs", 6) == 0)
		return (UNITSIZ / BLKSIZ) * (UNITMAX - UNITMIN + 1);
	else
		return 0;
}

int easy_dir(int fd, char *name) {
	entry	ent;
	int	unit, unitfd;

	if(check_fd(fd) == -1) return -1;
	unit = fd / (UNITSIZ / BLKSIZ);
	unitfd = fd % (UNITSIZ / BLKSIZ);
	rom_read((char*)(&ent), (DATATYPE)(unitfd * sizeof(entry) + unit * UNITSIZ), sizeof(entry));
	memcpy(name, ent.name, NAMESIZ);
	return ent.size;
}

int easy_find(char *name) {
	entry	ent;
	int	i, minfd, maxfd, n, unit;

	n = strlen(name);
	if(n > NAMESIZ) n = NAMESIZ;
	minfd = (UNITSIZ * sizeof(entry)) / (BLKSIZ * BLKSIZ);
	maxfd = UNITSIZ / BLKSIZ;
	for(unit = UNITMIN;unit <= UNITMAX;unit++) {
		for(i = minfd * sizeof(entry);i <= maxfd * sizeof(entry);i += sizeof(entry)) {
			rom_read((char*)(&ent), (DATATYPE)(i + unit * UNITSIZ), sizeof(entry));
			if(memcmp(ent.name, name, n) == 0) return i / sizeof(entry) + unit *maxfd;
		}
	}
	return -1;
}

int easy_del(char *name) {
	entry	ent;
	char	*buf;
	int	i, fd, n, unit, unitfd;

	fd = easy_find(name);
	if(fd == -1) return -1;
	unit = fd / (UNITSIZ / BLKSIZ);
	unitfd = fd % (UNITSIZ / BLKSIZ);
	buf = (char*)(&ent);
	rom_read(buf, (DATATYPE)(unitfd * sizeof(entry) + unit * UNITSIZ), sizeof(entry));
	n = ((ent.size - 1) / BLKSIZ) + 1;
#ifdef INTER_ROM
	bzero(buf, sizeof(entry));
#else
	for(i = 0;i < sizeof(entry);i++) buf[i] = 0xff;
#endif
	for(i = 0;i < n;i++) {
		rom_write((DATATYPE)((unitfd + i) * sizeof(entry) + unit * UNITSIZ), buf, sizeof(entry));
	}
	return fd;
}

int easy_create_unit(char *name, int size, int unit) {
	entry	ent;
	int	i, j, b, free, n, minfd, maxfd;

	minfd = (UNITSIZ * sizeof(entry)) / (BLKSIZ * BLKSIZ);
	maxfd = UNITSIZ / BLKSIZ;
	n = strlen(name);
	if(n > NAMESIZ) n = NAMESIZ;
	b = ((size - 1) / BLKSIZ) + 1;
	for(i = minfd * sizeof(entry);i <= (maxfd - b) * sizeof(entry);i += sizeof(entry)) {
		free = 0;
		for(j = 0;j < b * sizeof(entry);j += sizeof(entry)) {
			rom_read((char*)(&ent), (DATATYPE)(i + j + unit * UNITSIZ), sizeof(entry));
			if(ent.name[0] != 0xff) free = 1;
		}
		if(free == 0) {
			bzero((char*)(&ent), sizeof(entry));
			ent.size = size;
			bcopy(name, ent.name, n);
			rom_write((DATATYPE)(i + unit * UNITSIZ), (char*)(&ent), sizeof(entry));
			bzero((char*)(&ent), sizeof(entry));
			ent.name[0] = 0x7f;
			for(j = sizeof(entry);j < b * sizeof(entry);j += sizeof(entry)) {
				rom_write(i + j + unit * UNITSIZ + 4, ent.name, 1);
			}
			return i / sizeof(entry) + maxfd * unit;
		}
	}
	return -1;
}

int easy_create(char *name, int size) {
	int	unit, ret, count;

	easy_del(name);
	count = 0;
	for(unit = UNITMIN;unit <= UNITMAX;unit++) {
#ifdef INTER_ROM
		if(unit == UNITMAX) {
			if(count > 0) return 0;
			deflag();
			count++;
			unit = UNITMIN - 1;
			continue;
		}
#endif
		ret = easy_create_unit(name, size, unit);
		if(ret != -1) return ret;
	}
	return -1;
}

int easy_write(int fd, char *data, int offset, int size) {
	entry	ent;
	int	unit, unitfd;

	if(check_fd(fd) == -1) return -1;
	unit = fd / (UNITSIZ / BLKSIZ);
	unitfd = fd % (UNITSIZ / BLKSIZ);
	rom_read((char*)(&ent), (DATATYPE)(unitfd * sizeof(entry) + unit * UNITSIZ), sizeof(entry));
	if(offset + size > ent.size) return -1;
	rom_write((DATATYPE)(unitfd * BLKSIZ + offset + unit * UNITSIZ), data, size);
	return offset + size;
}

int easy_read(int fd, char *data, int offset, int size) {
	entry	ent;
	int	unit, unitfd;

	if(check_fd(fd) == -1) return -1;
	unit = fd / (UNITSIZ / BLKSIZ);
	unitfd = fd % (UNITSIZ / BLKSIZ);
	rom_read((char*)(&ent), (DATATYPE)(unitfd * sizeof(entry) + unit * UNITSIZ), sizeof(entry));
	if(offset + size > ent.size) return -1;
	rom_read(data, (DATATYPE)(unitfd * BLKSIZ + offset + unit * UNITSIZ), size);
	return offset + size;
}

int easy_exec(int argc, char **argv, char *user_area) {
	int	i, n, num, ret;
	char	*p;
	void	(*func)(int, char**);

	ret = easy_find(argv[0]);
	if(ret == -1) return -1;
	{
		char	buf[16];
		n = easy_dir(ret, buf);
	}
	p = (char*)user_area;
	for(i = 0;i < n;i += 32) {
		num = n - i;
		if(num > 32) num = 32;
		easy_read(ret, &p[i], i, num);
	}
	func = (void*)(user_area);
	(*func)(argc, argv);
	return 0;
}

#ifdef INTER_ROM
void copy_unit(int dest, int src) {
	char	buf[128];
	int	i;

	for(i = 0;i < UNITSIZ;i += 128) {
		rom_read(buf, (DATATYPE)(src * UNITSIZ + i), 128);
		rom_write((DATATYPE)(dest * UNITSIZ + i), buf, 128);
	}
	erase_unit(src);
}

void deflag() {
	entry	ent;
	int	i, eb, minfd, maxfd, n, unit, ret;

	minfd = (UNITSIZ * sizeof(entry)) / (BLKSIZ * BLKSIZ);
	maxfd = UNITSIZ / BLKSIZ;
	eb = UNITMIN;
	for(unit = UNITMIN;unit < UNITMAX;unit++) {
		for(i = minfd * sizeof(entry);i <= maxfd * sizeof(entry);i += sizeof(entry)) {
			rom_read((char*)(&ent), (DATATYPE)(i + unit * UNITSIZ), sizeof(entry));
			if(ent.name[0] != 0 && ent.name[0] != 0xff && ent.name[0] != 0x7f) {
				ret = easy_create_unit(ent.name, ent.size, UNITMAX);
				if(ret == -1) {
					erase_unit(eb);
					copy_unit(eb++, UNITMAX);
				}
			}
		}
	}
	erase_unit(eb);
	copy_unit(eb++, UNITMAX);
	for(;eb < unit;eb++) erase_unit(eb);
}
#endif
