#include "regdef.h"
#include "i2c.h"
#include "fs.h"

#define WRITE	0xa0
#define READ	0xa1

extern volatile REGTYPE*	i2c_regp[2];
extern volatile REGTYPE*	i2c_ddrp[2];
extern REGTYPE		i2c_bits[2][2];
int	eepsize, timeover;
void	i2c_start(int);
void	i2c_stop(int);
int	i2c_write(int, int);
int	i2c_read(int);
void	i2c_ack(int);
void	i2c_nak(int);
int	i2c_status(int);
int	send_ctrl(int, int);
#ifdef WITH_FS
  extern diskinfo	ee0;
#endif

int write_prepare(int addr) {
	int	ret;

	ret = send_ctrl(WRITE | ((addr >> 15) & 0x06), EEP);
	if(ret == -1) return -1;
	i2c_write((addr >> 8) & 0xff, EEP);
	i2c_write(addr & 0xff, EEP);
	return 0;
}

int read_prepare(int addr) {
	int	ret;

	ret = write_prepare(addr);
	if(ret == -1) return -1;
	ret = send_ctrl(READ | ((addr >> 15) & 0x06), EEP);
	return ret;
}

int eep_write(int dest, char *src, int size) {
	int	i, n, ret;

	if(size <= 0) return 0;
	if(size > 256) return -1;
	if(dest >= 0x40000) return -1;
	ret = write_prepare(dest);
	if(ret == -1) return -1;
	if((int)src == -1) for(i = 0;i < size;i++) i2c_write(0xff, EEP);
	else for(i = 0;i < size;i++) i2c_write(src[i], EEP);
	i2c_stop(EEP);
	return size;
}

int eep_read(char *dest, int src, int size) {
	int	i, n, ret;

	if(size == 0) return 0;
	if(size > 256) return -1;
	if(src >= 0x40000) return -1;
	ret = read_prepare(src);
	if(ret == -1) return -1;
	for(i = 0;i < size - 1;i++) {
		dest[i] = i2c_read(EEP);
		i2c_ack(EEP);
	}
	dest[size - 1] = i2c_read(EEP);
	i2c_nak(EEP);
	i2c_stop(EEP);
	return size;
}

#ifndef SH2
#ifdef H8_3664
static const int drlist[5] = {
  (int)&PDR1,(int)&PDR5,(int)&PDR7,(int)&PDR8,0
};
static const int ddrlist[5] = {
  (int)&PCR1,(int)&PCR5,(int)&PCR7,(int)&PCR8,0
};
#endif
#ifdef H8_300H
static const int drlist[11] = {
  (int)&P1DR, (int)&P2DR, (int)&P3DR, (int)&P4DR,
  (int)&P5DR, (int)&P6DR, (int)&P8DR, (int)&P9DR,
  (int)&PADR, (int)&PBDR, 0
};
static const int ddrlist[11] = {
  (int)&P1DDR, (int)&P2DDR, (int)&P3DDR, (int)&P4DDR,
  (int)&P5DDR, (int)&P6DDR, (int)&P8DDR, (int)&P9DDR,
  (int)&PADDR, (int)&PBDDR, 0
};
#endif

int	eep_probe() {
	char	check1, check2;
	int	ret, device, r, i, j;

	device = EEP;
	timeover = 60;
	for(r = 0;drlist[r] != 0;r++) {
		i2c_ddrp[device] = (volatile REGTYPE *)(ddrlist[r]);
		i2c_regp[device] = (volatile REGTYPE *)(drlist[r]);
		for(i = 1;i < 0x100;i <<= 1) {
			i2c_bits[0][device] = i;
			for(j = 1;j < 0x100;j <<= 1) {
				if(i == j) continue;
				i2c_bits[1][device] = j;
				I2CDDR = SCL;
				I2CDR = 0;
				if(I2CDR & SDA) {
					ret = send_ctrl(READ, EEP);
					i2c_stop(EEP);
					if(ret == 0) {
						I2CDDR = SCL;
						I2CDR = 0;
						timeover = 150;
						eepsize = eep_check();
						return 0;
					}
				}
			}
		}
		I2CDDR = 0;
	}
	i2c_ddrp[device] = 0;
	i2c_regp[device] = 0;
	eepsize = 0;
	return -1;
}
#endif

int	eep_check() {
	char	check1, check2, save;
	int	ret, mask, match;

	ret = eep_read(&save, 0, 1);
	if(ret == -1) return 0;
	for(mask = 0x10;mask < 0x10000;mask <<= 1) {
		match = 0;
		check1 = 0xa5, check2 = 0xff;
		eep_write(0, &check1, 1);
		eep_read(&check2, mask, 1);
		if(check1 == check2) match++;
		check1 = 0x5a, check2 = 0xff;
		eep_write(0, &check1, 1);
		eep_read(&check2, mask, 1);
		if(check1 == check2) match++;
		if(match == 2) break;
	}
	eep_write(0, &save, 1);
	if(eep_read(&save, 0x10000, 1) != -1) mask <<= 1;
	if(eep_read(&save, 0x20000, 1) != -1) mask <<= 1;
	return mask;
}

int get_eepsize() {
	return eepsize;
}

int eep_setup(volatile REGTYPE *drreg, volatile REGTYPE *ddrreg, REGTYPE *bits) {
	int	device;

	device = EEP;
	timeover = 150;
	i2c_bits[0][device] = bits[0];
	i2c_bits[1][device] = bits[1];
	i2c_ddrp[device] = ddrreg;
	i2c_regp[device] = drreg;
	I2CDDR &= ~SDA;
	I2CDDR |= SCL;
	I2CDR &= ~SCL;
	eepsize = eep_check();
	if(eepsize > 0) {
#ifdef WITH_FS
		get_diskinfo(EE0, &ee0);
#endif
	}
	return eepsize;
}
