#include	<string.h>
#include	<ctype.h>
#include	"fs.h"

int	start_dirent(char*, diskinfo*);
int	find_dirent_index(char*, char*, int, diskinfo*, int);
void	disp_dosname(char*, char*);
int	new_dir(char*, diskinfo*);
int	rm_file(char*, diskinfo*);
int	find_file(char*, fileinfo*, diskinfo*);
int	new_file(char*, fileinfo*, diskinfo*);
int	read_block(char*, fileinfo*, diskinfo*);
int	write_block(char*, int, fileinfo*, diskinfo*);
int	replace_block(char*, int, fileinfo*, diskinfo*);
int	rename_file(char*, char*, diskinfo*);
diskinfo* get_device(char*);
extern diskinfo	hd0, hd1, ee0;

#define	FD_OFF	8
fileinfo files[FD_MAX];
char	filebuf[FD_MAX][SECSIZE];
int	fileindex[FD_MAX];
int	filemax[FD_MAX];
int	fileopt[FD_MAX];
diskinfo* filedisk[FD_MAX];
int	file_reserve;

#define	FILEREAD	1
#define	FILEWRITE	2
#define	FILEAPPEND	3

typedef struct _dirent {
	unsigned int	fat;
	unsigned int	size;
	char		*path;
	diskinfo	*disk;
	unsigned short	max;
	unsigned short	index;
	unsigned char	attr;
} dirent;

int dirent_init(char *data, dirent *ent) {
	char	*name, entry[ENTSIZ];
	diskinfo *disk;

	if((disk = get_device(data)) == NULL) return -1;
	name = strchr(&(data[1]), '/');
	if(name == NULL) return -1;
	ent->fat = start_dirent(name, disk);
	if(ent->fat == -1) return -1;
	ent->max = find_dirent_index(entry, NULL, ent->fat, disk, -1);
	ent->disk = disk;
	ent->index = 0;
	return ent->max;
}

int dirent_get(dirent *ent) {
	char	entry[ENTSIZ], *ptr;

	if(ent->disk->device_id == 0) return -1;
	if(ent->index < ent->max)
		find_dirent_index(entry, ent->path, ent->fat, ent->disk, (ent->index)++);
	else
		return -1;
	if(ent->path[0] == 0) disp_dosname(ent->path, entry);
	ent->size = get_le_long(entry, FSIZE);
	ent->attr = entry[FATTR];
	return 0;
}

int makedir(char *data) {
	char	*name;
	diskinfo *disk;

	if((disk = get_device(data)) == NULL) return -1;
	name = strchr(&(data[1]), '/');
	if(name == NULL) return -1;
	return new_dir(name, disk);
}

int renfile(char *old, char *new) {
	char	*oldname, *newname;
	diskinfo *disk;
	int	id;
	
	if((disk = get_device(old)) == NULL) return -1;
	id = disk->device_id;
	if((disk = get_device(new)) == NULL) return -1;
	if(id != disk->device_id) return -1;
	oldname = strchr(&(old[1]), '/');
	if(oldname == NULL) return -1;
	newname = strchr(&(new[1]), '/');
	if(newname == NULL) return -1;
	return rename_file(oldname, newname, disk);
}

int delete(char *data) {
	char	*name;
	diskinfo *disk;

	if((disk = get_device(data)) == NULL) return -1;
	name = strchr(&(data[1]), '/');
	if(name == NULL) return -1;
	return rm_file(name, disk);
}

int fd_check(int _fd) {
	int	fd;

	fd = _fd - FD_OFF;
	if(fd < 0 || fd > FD_MAX) return -1;
	if(file_reserve & (1 << fd)) return -1;
	if(filedisk[fd]->device_id == 0) return -1;
	return fd;
}

int readfile(char *data, int _size, int _fd) {
	int	fd, size, n, s;

	if((fd = fd_check(_fd)) == -1) return -1;
	if(fileopt[fd] != FILEREAD) return -1;
	size = _size;

	while(size > 0) {
		n = filemax[fd] - fileindex[fd];
		if(n > size) s = size;
		else s = n;
		memcpy(&(data[_size - size]), &(filebuf[fd][fileindex[fd]]), s);
		fileindex[fd] += s;
		size -= s;
		if(size > 0) {
			filemax[fd] = read_block(filebuf[fd], &(files[fd]), filedisk[fd]);
			fileindex[fd] = 0;
			if(filemax[fd] == 0) return (_size - size);
		}
	}
	return (_size - size);
}

int writefile(char *data, int _size, int _fd) {
	int	fd, size, eof, n, s;

	if((fd = fd_check(_fd)) == -1) return -1;
	if(fileopt[fd] != FILEWRITE && fileopt[fd] != FILEAPPEND) return -1;
	size = _size;

	while(size > 0) {
		n = (filedisk[fd])->sector_size - fileindex[fd];
		if(n > size) s = size;
		else s = n;
		memcpy(&(filebuf[fd][fileindex[fd]]), &(data[_size - size]), s);
		fileindex[fd] += s;
		size -= s;
		if(fileindex[fd] >= (filedisk[fd])->sector_size) {
			if(fileopt[fd] == FILEAPPEND)
				replace_block(filebuf[fd], fileindex[fd], &(files[fd]), filedisk[fd]);
			else
				write_block(filebuf[fd], fileindex[fd], &(files[fd]), filedisk[fd]);
			fileopt[fd] = FILEWRITE;
			fileindex[fd] = 0;
		}
	}
	return _size;
}

int openfile(char *data, int opt) {
	int	fd, m, ret;
	char	*name;

	for(m = 1,fd = 0;fd < FD_MAX;m <<= 1,fd++) if(file_reserve & m) break;
	if(fd == FD_MAX) return -1;
	if((filedisk[fd] = get_device(data)) == NULL) return -1;
	name = strchr(&(data[1]), '/');
	if(name == NULL) return -1;
	ret = find_file(name, &(files[fd]), filedisk[fd]);
	fileopt[fd] = opt;
	filemax[fd] = fileindex[fd] = 0;
	if(fileopt[fd] == FILEAPPEND && ret == -1) fileopt[fd] = FILEWRITE;
	switch(fileopt[fd]) {
	case FILEREAD:
		if(ret == -1) return -1;
		break;
	case FILEAPPEND:
		while((fileindex[fd] = read_block(filebuf[fd], &(files[fd]), filedisk[fd])) == SECSIZE);
		break;
	case FILEWRITE:
		if(ret != -1) {
			if(ret & FATTR_DIR) return -1;
			rm_file(name, filedisk[fd]);
		}
		ret = new_file(name, &(files[fd]), filedisk[fd]);
		if(ret == -1) return -1;
		break;
	default:
		return -1;
	}
	file_reserve &= ~(1 << fd);
	return fd + FD_OFF;
}

int closefile(int _fd) {
	int	fd;

	if((fd = fd_check(_fd)) == -1) return -1;
	switch(fileopt[fd]) {
	case FILEWRITE:
		write_block(filebuf[fd], fileindex[fd], &(files[fd]), filedisk[fd]);
		break;
	case FILEAPPEND:
		replace_block(filebuf[fd], fileindex[fd], &(files[fd]), filedisk[fd]);
		break;
	}
	file_reserve |= 1 << fd;
	return 0;
}

int get_file_size(int _fd) {
	int	fd;

	if((fd = fd_check(_fd)) == -1) return -1;
	return files[fd].size;
}

void file_init() {
	file_reserve = 0x07;
}
