#include	<string.h>
#include	<ctype.h>
#include	"fs.h"

extern unsigned char	sector_buffer[SECSIZE];
int			sector_write(unsigned int, diskinfo*);
int			sector_read(unsigned int, diskinfo*);
int			mbr_read(int, int);

unsigned char checksum(char *fname){
	unsigned char	sum, rotate;
	int		i;

	sum = 0;
	for(i = 0;i < 11;++i){
		if(1 & sum) rotate = 0x80;
		else rotate = 0;
		sum = sum >> 1;
		sum = sum | rotate;
		sum += fname[i];
	}
	return(sum);
}

void	set_word(unsigned short data, unsigned int offset) {
	sector_buffer[offset] = data;
	sector_buffer[offset + 1] = data >> 8;
}

unsigned short	get_word(unsigned int offset) {
	unsigned short	data;

	data = sector_buffer[offset] & 0xff;
	data += (sector_buffer[offset + 1] & 0xff) << 8;
	return data;
}

void	set_le_word(unsigned short data, char *buffer, unsigned int offset) {
	buffer[offset] = data;
	buffer[offset + 1] = data >> 8;
}

unsigned short	get_le_word(char *buffer, unsigned int offset) {
	unsigned short	data;

	data = buffer[offset] & 0xff;
	data += (buffer[offset + 1] & 0xff) << 8;
	return data;
}

void	set_long(unsigned int data, unsigned int offset) {
	sector_buffer[offset] = data;
	sector_buffer[offset + 1] = data >> 8;
	sector_buffer[offset + 2] = data >> 16;
	sector_buffer[offset + 3] = data >> 24;
}

unsigned int	get_long(unsigned int offset) {
	unsigned int	data;

	data = sector_buffer[offset] & 0xff;
	data += (sector_buffer[offset + 1] & 0xff) << 8;
	data += (sector_buffer[offset + 2] & 0xff) << 16;
	data += (sector_buffer[offset + 3] & 0xff) << 24;
	return data;
}

void	set_le_long(unsigned int data, char *buffer, unsigned int offset) {
	buffer[offset] = data;
	buffer[offset + 1] = data >> 8;
	buffer[offset + 2] = data >> 16;
	buffer[offset + 3] = data >> 24;
}

unsigned int	get_le_long(char *buffer, unsigned int offset) {
	unsigned int	data;

	data = buffer[offset] & 0xff;
	data += (buffer[offset + 1] & 0xff) << 8;
	data += (buffer[offset + 2] & 0xff) << 16;
	data += (buffer[offset + 3] & 0xff) << 24;
	return data;
}

int	set_dosname(char *dest, char *src) {
	char	*ptr;
	int	i, e, d;

	for(i = 0;i < 11;i++) dest[i] = ' ';
	dest[i] = 0;
	ptr = strchr(src, '.');
	if(ptr == NULL) {
		d = strlen(src); e = 0;
	} else {
		d = (int)ptr - (int)src; e = strlen(ptr);
		memcpy(&(dest[8]), &(ptr[1]), (e > 3) ? 3 : e);
	}
	if(d <= 8) {
		memcpy(dest, src, d);
	} else {
		memcpy(dest, src, 6);
		dest[6] = '~'; dest[7] = '0';
	}
	for(i = 0;i < 11;i++)
		if(dest[i] >= 'a' && dest[i] <= 'z')
			dest[i] = toupper(dest[i]);
	return d;
}

void disp_dosname(char *dst, char *src) {
	char	*ptr;
	int	n;

	bzero(dst, 13);
	memcpy(dst, src, 8);
	ptr = strchr(dst, ' ');
	if(ptr == NULL) ptr = &(dst[8]);
	*(ptr++) = '.';
	memcpy(ptr, &(src[8]), 3);
	ptr = strchr(ptr, ' ');
	if(ptr != NULL) *ptr = 0;
	n = strlen(dst);
	if(dst[n - 1] == '.' && n > 0) dst[n - 1] = 0;
}

unsigned int	fat2sector(unsigned int clustor, diskinfo *info) {
	unsigned int	rde, sector;

	sector = 0;
	if(info->fat == FAT12 || info->fat == FAT16) {
		rde = info->root_dir_entry * ENTSIZ;
		sector = rde / info->sector_size;
		if((rde % info->sector_size) != 0) sector++;
	}
	sector += info->reserve_sector + info->fat_sector_size * info->fat_number;
	sector += (clustor - 2) * info->sector_per_clustor;
	return sector;
}

unsigned int	next_fat(unsigned int clustor, diskinfo *info) {
	unsigned int	addr, fats;

	if(info->fat == FAT12) {
		addr = clustor >> 1;
		addr = addr << 1;
		addr = (addr * 3) / 2;
		sector_read(addr / info->sector_size + info->reserve_sector, info);
		addr = addr % info->sector_size;
		fats = sector_buffer[addr] & 0xff;
		fats += (sector_buffer[addr + 1] & 0xff) << 8;
		fats += (sector_buffer[addr + 2] & 0xff) << 16;
		if(clustor & 1) {
			fats &= 0xfff000;
			fats >>= 12;
		} else {
			fats &= 0x000fff;
		}
	} else if(info->fat == FAT16) {
		addr = clustor * 2;
		sector_read(addr / info->sector_size + info->reserve_sector, info);
		addr = addr % info->sector_size;
		fats = get_word(addr);
	} else if(info->fat == FAT32) {
		addr = clustor * 4;
		sector_read(addr / info->sector_size + info->reserve_sector, info);
		addr = addr % info->sector_size;
		fats = get_long(addr);
	}
	return fats;
}

void	set_fat(unsigned int clustor, unsigned int next, diskinfo *info) {
	unsigned int	addr, fats;

	if(info->fat == FAT12) {
		addr = clustor >> 1;
		addr = addr << 1;
		addr = (addr * 3) / 2;
		sector_read(addr / info->sector_size + info->reserve_sector, info);
		addr = addr % info->sector_size;
		fats = sector_buffer[addr] & 0xff;
		fats += (sector_buffer[addr + 1] & 0xff) << 8;
		fats += (sector_buffer[addr + 2] & 0xff) << 16;
		if(clustor & 1) {
			fats &= 0x000fff;
			fats |= (next & 0xfff) << 12;
		} else {
			fats &= 0xfff000;
			fats |= next & 0xfff;
		}
		sector_buffer[addr] = fats;
		sector_buffer[addr + 1] = fats >> 8;
		sector_buffer[addr + 2] = fats >> 16;
		addr = clustor >> 1;
		addr = addr << 1;
		addr = (addr * 3) / 2;
	} else if(info->fat == FAT16) {
		addr = clustor * 2;
		sector_read(addr / info->sector_size + info->reserve_sector, info);
		addr = addr % info->sector_size;
		set_word(next, addr);
	} else if(info->fat == FAT32) {
		addr = clustor * 4;
		sector_read(addr / info->sector_size + info->reserve_sector, info);
		addr = addr % info->sector_size;
		set_long(next, addr);
	}
	sector_write(addr / info->sector_size + info->reserve_sector, info);
	sector_write(addr / info->sector_size + info->reserve_sector + info->fat_sector_size, info);
}

unsigned int	get_fat(diskinfo *info) {
	unsigned int	i, index, fats, count, addr, max_sec, rde, sec;

	rde = info->root_dir_entry * ENTSIZ;
	sec = rde / info->sector_size;
	if((rde % info->sector_size) != 0) sec++;
	sec += info->reserve_sector + info->fat_sector_size * info->fat_number;
	max_sec = info->media_sector - sec;
	if(info->fat == FAT12) {
		count = index = 0;
		for(i = 0;i < (max_sec / (info->sector_per_clustor * 2));i++) {
			addr = index % info->sector_size;
			if(addr == 0) sector_read(index / info->sector_size + info->reserve_sector, info);
			fats = sector_buffer[addr] & 0xff;
			fats += (sector_buffer[addr + 1] & 0xff) << 8;
			fats += (sector_buffer[addr + 2] & 0xff) << 16;
			if((fats & 0xfff) == 0) return count;
			if((fats & 0xfff000) == 0) return count + 1;
			count += 2;
			index += 3;
		}
	} else if(info->fat == FAT16) {
		index = 0;
		for(count = 0;count < (max_sec / info->sector_per_clustor);count++) {
			addr = index % info->sector_size;
			if(addr == 0) sector_read(index / info->sector_size + info->reserve_sector, info);
			if(get_word(addr) == 0) return count;
			index += 2;
		}
	} else if(info->fat == FAT32) {
		index = 0;
		for(count = 0;count < (max_sec / info->sector_per_clustor);count++) {
			addr = index % info->sector_size;
			if(addr == 0) sector_read(index / info->sector_size + info->reserve_sector, info);
			if(get_long(addr) == 0) return count;
			index += 4;
		}
	}
	return -1;
}

unsigned int	rm_fat(unsigned int clustor, diskinfo *info) {
	unsigned int	fats, next, max;

	if(info->fat == FAT12) max = 0xff8;
	else if(info->fat == FAT16) max = 0xfff8;
	else if(info->fat == FAT32) max = 0xffffff8;
	fats = clustor;
	while(fats < max) {
		next = next_fat(fats, info);
		set_fat(fats, 0, info);
		fats = next;
	}
}

unsigned int get_dirent_top(unsigned int dirfat, diskinfo *info) {
	if(dirfat == 0)
		return info->reserve_sector + info->fat_sector_size * info->fat_number;
	else
		return fat2sector(dirfat, info);
}

unsigned int get_dirent_next(unsigned int index, unsigned int entsec, unsigned int *dirfat, diskinfo *info) {
	unsigned int	fat_siz, sec_off, max;

	if(info->fat == FAT12) max = 0xff8;
	else if(info->fat == FAT16) max = 0xfff8;
	else if(info->fat == FAT32) max = 0xffffff8;
	fat_siz = info->sector_size * info->sector_per_clustor;
	sec_off = (index % fat_siz) / info->sector_size;
	if(*dirfat == 0) {
		sector_read(index / info->sector_size + entsec, info);
	} else {
		if((index % fat_siz) == 0 && index != 0) {
			*dirfat = next_fat(*dirfat, info);
			if(*dirfat >= max) return -1;
			entsec = fat2sector(*dirfat, info);
		}
		sector_read(sec_off + entsec, info);
	}
	return entsec;
}

int	get_dirent_index(unsigned int dirfat, int pathsize, diskinfo *info) {
	unsigned int	i, j, index, addr, entsec, entcnt, first, count;
	unsigned int	dir_fat, cur_fat, new_fat, ret, sec;
	unsigned char	*entry;

	entcnt = (pathsize - 1) / 13 + 2;
	first = count = index = 0;
	dir_fat = dirfat;
	entsec = get_dirent_top(dirfat, info);
	for(i = 0;;i++) {
		addr = index % info->sector_size;
		cur_fat = dir_fat;
		if(addr == 0) {
			ret = get_dirent_next(index, entsec, &dir_fat, info);
			if(ret == -1) {
				if(dirfat == 0) return -1;
				new_fat = get_fat(info);
				if(new_fat == -1) return -1;
				set_fat(cur_fat, new_fat, info);
				set_fat(new_fat, -1, info);
				sec = fat2sector(new_fat, info);
				bzero(sector_buffer, info->sector_size);
				for(j = 0;j < info->sector_per_clustor;j++) {
					sector_write(sec + j, info);
				}
				dir_fat = cur_fat;
				ret = get_dirent_next(index, entsec, &dir_fat, info);
			}
			entsec = ret;
		}
		entry = &(sector_buffer[addr]);
		if(entry[FNAME] == 0 || entry[FNAME] == 0xe5) {
			if((first + count) != index) {
				first = index;
				count = 0;
			}
			count += ENTSIZ;
		}
		if(count >= (entcnt * ENTSIZ)) return first; 
		index += ENTSIZ;
		if(dir_fat == 0 && i >= info->root_dir_entry - 1) return -1;
	}
	return -1;
}

int	find_dirent_index(char *dest, char *name, unsigned int dirfat, diskinfo *info, int pos) {
	unsigned int	i, j, index, first, addr, entsec, count;
	unsigned int	dir_fat, fat_siz, cur;
	unsigned char	*entry, *src, *dst, dosname[14], path[PATHSIZ];

	count = first = index = 0;
	dir_fat = dirfat;
	fat_siz = info->sector_size * info->sector_per_clustor;
	entsec = get_dirent_top(dir_fat, info);
	cur = 0;
	path[0] = 0;
	for(i = 0;;i++) {
		addr = index % info->sector_size;
		if(addr == 0) entsec = get_dirent_next(index, entsec, &dir_fat, info);
		if(entsec == -1) break;
		entry = &(sector_buffer[addr]);
		if(entry[FNAME] == 0 || entry[FNAME] == 0xe5) {
			first = index + ENTSIZ;
		} else if(entry[FATTR] == 0x0f) {
			cur = (((int)(entry[FNAME]) & 0x3f) - 1) * 13;
			if(cur < PATHSIZ) {
				src = &(entry[1]);
				dst = &(path[cur]);
				for(j = 0;j < 13;j++) {
					if(j == 5) src += 3;
					if(j == 11) src += 2;
					*(dst++) = *src;
					src += 2;
				}
			}
		} else if(cur == 0) {
			if(pos == -1 && name != NULL) {
				disp_dosname(dosname, entry);
				if(strcmp(path, name) == 0 ||
				   strcmp(dosname, name) == 0) {
					memcpy(dest, entry, ENTSIZ);
					return first;
				}
			} else if(count == pos && name != NULL) {
				memcpy(dest, entry, ENTSIZ);
				if(name != NULL) strncpy(name, path, PATHSIZ);
				return first;
			}
			path[0] = 0;
			first = index + ENTSIZ;
			count++;
		}
		index += ENTSIZ;
		if(dir_fat == 0 && i >= info->root_dir_entry - 1) break;
	}
	if(name == NULL) return count;
	else return -1;
}

int	set_dirent_index(unsigned int index, char *path, char *src, unsigned int dirfat, diskinfo *info) {
	unsigned int	j, n, addr, entsec, entcnt, off;
	unsigned int	dir_fat;
	int		pindex, i;
	unsigned char	*entry, sum;

	sum = checksum(src);
	entcnt = (strlen(path) - 1) / 13 + 2;
	dir_fat = dirfat;
	if(dirfat == 0) {
		entsec = index / info->sector_size + info->reserve_sector + info->fat_sector_size * info->fat_number;
	} else {
		n = index / (info->sector_size * info->sector_per_clustor);
		off = index;
		for(i = 0;i < n;i++) {
			dir_fat = next_fat(dir_fat, info);
			off -= info->sector_size * info->sector_per_clustor;
		}
		entsec = fat2sector(dir_fat, info);
		entsec += off / info->sector_size;
	}
	if(entsec == -1) return -1;
	sector_read(entsec, info);
	for(i = index;i < index + entcnt * ENTSIZ;i += ENTSIZ) {
		addr = i % info->sector_size;
		if(addr == 0) {
			sector_write(entsec, info);
			entsec = get_dirent_next(i, entsec, &dir_fat, info);
			if(entsec == -1) return -1;
		}
		entry = &(sector_buffer[addr]);
		if(entry[FNAME] == 0 || entry[FNAME] == 0xe5) {
			bzero(entry, 32);
			if(i == index) entry[0] = 0x40;
			pindex = (index + (entcnt - 2) * ENTSIZ) - i;
			pindex = (pindex / ENTSIZ) * 13;
			if(pindex >= 0) {
				entry[0] |= (pindex / 13) + 1;
				entry[FATTR] = 0x0f;
				entry[13] = sum;
				for(j = 0;j < 5;j++) entry[j * 2 + 1] = path[pindex + j];
				for(j = 0;j < 6;j++) entry[j * 2 + 14] = path[pindex + j + 5];
				for(j = 0;j < 2;j++) entry[j * 2 + 0x1c] = path[pindex + j + 11];
			} else {
				memcpy(entry, src, ENTSIZ);
				break;
			}
		} else {
			pindex = (index + (entcnt - 2) * ENTSIZ) - i;
			pindex = (pindex / ENTSIZ) * 13;
			if(pindex < 0) {
				memcpy(entry, src, ENTSIZ);
				break;
			}
		}
	}
	sector_write(entsec, info);
	return 0;
}

int	fileinfo_check(fileinfo *file) {
	if(file->name[0] == 0) return -1;
	if(file->top_fat == 0) return -1;
	if(file->size == 0) return -1;
	return 0;
}

int	match_dosname_count(unsigned char *name, unsigned int dirfat, diskinfo *info) {
	unsigned int	i, n, count;
	unsigned char	entry[ENTSIZ];

	count = 0;
	n = find_dirent_index(entry, NULL, dirfat, info, -1); 
	for(i = 0;i < n;i++) {
		find_dirent_index(entry, NULL, dirfat, info, i);
		entry[7] = '0';
		if(memcmp(entry, name, 11) == 0) count++;
	}
	return count;
}

int	set_dirent(fileinfo *file, diskinfo *info) {
	char	entry[ENTSIZ], dosname[12];
	int	index, n;

	if(fileinfo_check(file) == -1) return -1;
	bzero(entry, ENTSIZ);
	n = set_dosname(dosname, file->name);
	if(n > 8) {
		n = match_dosname_count(dosname, file->dir_fat, info);
		if(n < 9) dosname[7] = '1' + n;
		else if(n < 35) dosname[7] = 'A' + n - 9;
	}
	memcpy(&(entry[FNAME]), dosname, 11);
	entry[FATTR] = 0x20;
	set_le_word(file->top_fat, entry, FCLUSTOR);
	if(info->fat == FAT32) set_le_word((file->top_fat >> 16), entry, FCLUSTOR2);
	set_le_long(file->size, entry, FSIZE);
	index = file->dirent;
	if(file->dirent == -1) index = get_dirent_index(file->dir_fat, strlen(file->name), info);
	if(index == -1) return -1;
	file->dirent = index;
	set_dirent_index(index, file->name, entry, file->dir_fat, info);
	return 0;
}

int	rm_dirent(char *name, unsigned int dirfat, diskinfo *info) {
	unsigned int	addr, i, index, dir_fat, entsec, off, n, count, fat;
	char		*entry, _entry[ENTSIZ];

	index = find_dirent_index(_entry, name, dirfat, info, -1);
	if(index == -1) return -1;
	if(_entry[FATTR] & FATTR_DIR) {
		dir_fat = get_le_word(_entry,FCLUSTOR);
		if(info->fat == FAT32) dir_fat += get_le_word(_entry,FCLUSTOR2) << 16;
		count = find_dirent_index(_entry, NULL, dir_fat, info, -1);
		if(count > 2) return -2;
	}
	dir_fat = dirfat;
	if(dirfat == 0) {
		entsec = index / info->sector_size + info->reserve_sector + info->fat_sector_size * info->fat_number;
	} else {
		n = index / (info->sector_size * info->sector_per_clustor);
		off = index;
		for(i = 0;i < n;i++) {
			dir_fat = next_fat(dir_fat, info);
			off -= info->sector_size * info->sector_per_clustor;
		}
		entsec = fat2sector(dir_fat, info);
		entsec += off / info->sector_size;
	}
	if(entsec == -1) return -1;
	sector_read(entsec, info);
	for(i = index;;i += ENTSIZ) {
		addr = i % info->sector_size;
		if(addr == 0) {
			sector_write(entsec, info);
			entsec = get_dirent_next(i, entsec, &dir_fat, info);
			if(entsec == -1) return -1;
		}
		entry = &(sector_buffer[addr]);
		entry[FNAME] = 0xe5;
		if(entry[FATTR] != 0x0f || get_le_word(entry, FCLUSTOR) != 0) break;
	}
	sector_write(entsec, info);
	fat = get_le_word(entry, FCLUSTOR);
	if(info->fat == FAT32) fat += get_le_word(entry,FCLUSTOR2) << 16;
	return fat;
}

unsigned int get_dirfat(char *path, diskinfo *info, char *entry) {
	char	buffer[PATHSIZ];
	char	*ptr;
	unsigned int	index, dir_fat, ret;

	dir_fat = 0;
	if(info->fat == FAT32) dir_fat = info->root_dir_entry;
	buffer[PATHSIZ - 1] = 0;
	ptr = &(path[1]);
	if(*path != '/') return -1;
	if(*ptr == 0) return dir_fat;
	index = 0;
	while(*ptr != 0) {
		if(*ptr == '/') {
			buffer[index] = 0;
			index = 0;
			ret = find_dirent_index(entry, buffer, dir_fat, info, -1);
			if(ret == -1) return -1;
			if(!(entry[FATTR] & FATTR_DIR)) return -1;
			dir_fat = get_le_word(entry, FCLUSTOR);
			if(info->fat == FAT32) dir_fat += get_le_word(entry,FCLUSTOR2) << 16;
		} else if(index < (PATHSIZ - 1)) {
			buffer[index++] = *ptr; 
		}
		ptr++;
	}
	return dir_fat;
}

int	new_file(char *path, fileinfo *file, diskinfo *info) {
	char	*name, entry[ENTSIZ];

	file->dir_fat = get_dirfat(path, info, entry);
	if(file->dir_fat == -1) return -1;
	name = strrchr(path, '/');
	name++;
	file->top_fat = 0;
	file->cur_fat = 0;
	file->sec_off = 0;
	file->size = 0;
	file->cur_size = 0;
	file->dirent = -1;
	strncpy(file->name, name, PATHSIZ - 1);
	if(get_dirent_index(file->dir_fat, strlen(name), info) == -1) return -1;
	return 0;
}

void	make_entry(char *entry, char *filename, unsigned int fat, diskinfo *info) {
	bzero(entry, ENTSIZ);
	memcpy(&(entry[FNAME]), filename, 11);
	entry[FATTR] = FATTR_DIR;
	set_le_word(fat, entry, FCLUSTOR);
	if(info->fat == FAT32) set_le_word((fat >> 16), entry, FCLUSTOR2);
}

int	new_dir(char *path, diskinfo *info) {
	char		filename[12], *name;
	char		parent[ENTSIZ], entry[ENTSIZ];
	unsigned int	i, n, fat, dir_fat, index, sec;

	dir_fat = get_dirfat(path, info, parent);
	if(dir_fat == -1) return -1;
	name = strrchr(path, '/');
	name++;
	n = set_dosname(filename, name);
	if(n > 8) {
		n = match_dosname_count(filename, dir_fat, info);
		if(n < 9) filename[7] = '1' + n;
		else if(n < 35) filename[7] = 'A' + n - 9;
	}
	index = get_dirent_index(dir_fat, strlen(name), info);
	if(index == -1) return -1;
	fat = get_fat(info);
	if(fat == -1) return -1;
	if(dir_fat == 0) make_entry(parent, filename, 0, info);
	make_entry(entry, filename, fat, info);
	set_dirent_index(index, name, entry, dir_fat, info);
	set_fat(fat, -1, info);
	sec = fat2sector(fat, info);
	bzero(sector_buffer, info->sector_size);
	memcpy(sector_buffer, entry, ENTSIZ);
	memcpy(sector_buffer, ".          ", 11);
	memcpy(sector_buffer + ENTSIZ, parent, ENTSIZ);
	memcpy(sector_buffer + ENTSIZ, "..         ", 11);
	sector_write(sec, info);
	bzero(sector_buffer, info->sector_size);
	for(i = 1;i < info->sector_per_clustor;i++) {
		sector_write(sec + i, info);
	}
	return 0;
}

int	write_block(char *data, int size, fileinfo *file, diskinfo *info) {
	unsigned int fat, sec;

	if(size <= 0 && size > info->sector_size) return -1;
	if(file->sec_off == 0 && (file->size % info->sector_size) == 0) {
		fat = get_fat(info);
		if(fat == -1) return -1;
		if(file->top_fat == 0) {
			file->top_fat = fat;
		} else {
			set_fat(file->cur_fat, fat, info);
		}
		set_fat(fat, -1, info);
		file->cur_fat = fat;
	}
	sec = fat2sector(file->cur_fat, info) + file->sec_off;
	bzero(sector_buffer, info->sector_size);
	memcpy(sector_buffer, data, size);
	sector_write(sec, info);
	file->size = file->size / info->sector_size;
	file->size = file->size * info->sector_size;
	file->size += size;
	if(set_dirent(file, info) == -1) return -1;
	if(size == info->sector_size) {
		(file->sec_off)++;
		if(file->sec_off >= info->sector_per_clustor) file->sec_off = 0;
	}
	return 0;
}

int	replace_block(char *data, int size, fileinfo *file, diskinfo *info) {
	unsigned int fat, sec;

	if(size <= 0 && size > info->sector_size) return -1;
	sec = fat2sector(file->cur_fat, info) + file->sec_off;
	bzero(sector_buffer, info->sector_size);
	memcpy(sector_buffer, data, size);
	sector_write(sec, info);
	file->size = file->size / info->sector_size;
	file->size = file->size * info->sector_size;
	file->size += size;
	if(set_dirent(file, info) == -1) return -1;
	if(size == info->sector_size) {
		(file->sec_off)++;
		if(file->sec_off >= info->sector_per_clustor) file->sec_off = 0;
	}
	return 0;
}

int	read_block(char *data, fileinfo *file, diskinfo *info) {
	int	size, sec;

	size = file->size - file->cur_size;
	if(size == 0) return 0;
	if(size > info->sector_size) size = info->sector_size;
	file->cur_size += size;
	sec = fat2sector(file->cur_fat, info) + file->sec_off;
	sector_read(sec, info);
	memcpy(data, sector_buffer, size);
	if(size == info->sector_size) {
		(file->sec_off)++;
		if(file->sec_off >= info->sector_per_clustor) {
			file->sec_off = 0;
			file->cur_fat = next_fat(file->cur_fat, info);
		}
	}
	return size;
}

int	rename_file(char *old, char *new, diskinfo *info) {
	int	dir_fat1, dir_fat2, index1, index2, n;
	char	entry[ENTSIZ], *name1, *name2;

	dir_fat1 = get_dirfat(old, info, entry);
	if(dir_fat1 == -1) return -1;
	dir_fat2 = get_dirfat(new, info, entry);
	if(dir_fat2 == -1) return -1;
	name1 = strrchr(old, '/');
	name1++;
	name2 = strrchr(new, '/');
	name2++;
	index1 = find_dirent_index(entry, name1, dir_fat1, info, -1);
	if(index1 == -1) return -1;
	index2 = get_dirent_index(dir_fat2, strlen(name2), info);
	if(index2 == -1) return -1;
	n = set_dosname(entry, name2);
	if(n > 8) {
		n = match_dosname_count(entry, dir_fat2, info);
		if(n < 9) entry[7] = '1' + n;
		else if(n < 35) entry[7] = 'A' + n - 9;
	}
	set_dirent_index(index2, name2, entry, dir_fat2, info);
	rm_dirent(name1, dir_fat1, info);
	return 0;
}

int	rm_file(char *path, diskinfo *info) {
	int	fat, dir_fat;
	char	entry[ENTSIZ], *name;

	dir_fat = get_dirfat(path, info, entry);
	if(dir_fat == -1) return -1;
	name = strrchr(path, '/');
	name++;
	fat = rm_dirent(name, dir_fat, info);
	if(fat < 0) return fat;
	rm_fat(fat, info);
	return 0;
}

int	find_file(char *path, fileinfo *file, diskinfo *info) {
	int	index, i;
	char	entry[ENTSIZ], *name;

	file->dir_fat = get_dirfat(path, info, entry);
	if(file->dir_fat == -1) return -1;
	name = strrchr(path, '/');
	name++;
	strncpy(file->name, name, PATHSIZ - 1);
	index = find_dirent_index(entry, file->name, file->dir_fat, info, -1);
	if(index == -1) return -1;
	file->top_fat = get_le_word(entry, FCLUSTOR);
	if(info->fat == FAT32) file->top_fat += get_le_word(entry,FCLUSTOR2) << 16;
	file->cur_fat = file->top_fat;
	file->size = get_le_long(entry, FSIZE);
	file->sec_off = 0;
	file->cur_size = 0;
	file->dirent = index;
	return entry[FATTR] & 0xff;
}

int	start_dirent(char *path, diskinfo *info) {
	int	dir_fat;
	char	entry[ENTSIZ], *name;

	dir_fat = get_dirfat(path, info, entry);
	if(dir_fat == -1) return -1;
	return dir_fat;
}

int	get_diskinfo(int id, diskinfo *info) {
	info->device_id = 0;
	info->sector_offset = 0;
	if(mbr_read(id, 0) == -1) return -1;
	info->fat = NOFAT;
	if(sector_buffer[0] != 0xeb ||
	   sector_buffer[0x1fe] != 0x55 ||
	   sector_buffer[0x1ff] != 0xaa) {
		info->sector_offset = get_long(0x1c6);
		if(mbr_read(id, info->sector_offset) == -1) return -1;
		if(sector_buffer[0] != 0xeb ||
		   sector_buffer[0x1fe] != 0x55 ||
		   sector_buffer[0x1ff] != 0xaa) {
			return -1;
		}
	}
	if(memcmp(&(sector_buffer[SIG]), "FAT12   ", 8) == 0) info->fat = FAT12;
	if(memcmp(&(sector_buffer[SIG]), "FAT16   ", 8) == 0) info->fat = FAT16;
	if(memcmp(&(sector_buffer[SIG32]), "FAT32   ", 8) == 0) info->fat = FAT32;
	if(info->fat == NOFAT) {
		if(get_word(RDE) == 0) {
			if(get_long(RDSC) != 0) info->fat = FAT32;
		} else {
			if(get_word(TS) < 0x8000) info->fat = FAT12;
			else info->fat = FAT16;
		}
	}
	info->sector_size = get_word(SS);
	info->sector_per_clustor = sector_buffer[SC];
	info->reserve_sector = get_word(RSC);
	info->fat_number = sector_buffer[FN];
	info->media_ident = sector_buffer[MEDIA_IDENT];
	switch(info->fat) {
	case FAT12:
	case FAT16:
		info->root_dir_entry = get_word(RDE);
		info->fat_sector_size = get_word(SF);
		info->media_sector = get_word(TS);
		if(info->media_sector == 0) info->media_sector = get_long(TS32);
		break;
	case FAT32:
		info->root_dir_entry = get_long(RDSC);
		info->media_sector = get_long(TS32);
		info->fat_sector_size = get_long(SF32);
		break;
	}
	info->device_id = id;
	return 0;
}
