#define	RS	masks[0]
#define	RW	masks[1]
#define	CS	masks[2]
#define	D7	masks[3]
#define	D6	masks[4]
#define	D5	masks[5]
#define	D4	masks[6]
#define	D3	masks[3]
#define	D2	masks[4]
#define	D1	masks[5]
#define	D0	masks[6]
#define DATA	masks164[0]
#define CLK	masks164[1]
#define ENA	masks164[2]
#define	NONE	0

#define LCDINC		D1
#define LCDSHIFT	D0
#define LCDDISPLAY	D2
#define LCDCURSOR	D1
#define LCDBLINK	D0
#define LCDDISPSHIFT	D3
#define LCDRIGHT	D2
#define LCDLEFT		NONE

#ifdef	SH2
  #define WAITSCALE	3
#else
  #define WAITSCALE	1
#endif

static unsigned short current, max_low, max_column;
static unsigned REGTYPE masks[7];
static unsigned REGTYPE masks164[3];
static volatile unsigned REGTYPE *lcdreg;
static char databuf[2][40];
#define SerialMode	masks164[0]

void serial_out(int data) {
	int	i;

	*lcdreg &= ~(DATA | CLK);
	for(i = 0;i < 8;i++) {
		if(data & 0x80) *lcdreg |= DATA;
		else *lcdreg &= ~DATA;
		*lcdreg |= CLK;
		*lcdreg &= ~CLK;
		data <<= 1;
	}
}

void outlcd(int data) {
	volatile int	i;

	if(SerialMode) {
		serial_out(data);
		*lcdreg |= ENA;
		for(i = 0;i < 10*WAITSCALE;i++);
		*lcdreg &= ~ENA;
		for(i = 0;i < 80*WAITSCALE;i++);
	} else {
		*lcdreg = CS | data;
		for(i = 0;i < 10*WAITSCALE;i++);
		*lcdreg = data;
		for(i = 0;i < 80*WAITSCALE;i++);
	}
}

void Clear() {
	volatile int	i;

	outlcd(NONE); outlcd(D0);
	for(i = 0;i < 3600*WAITSCALE;i++);	
}

void Home() {
	volatile int	i;

	outlcd(NONE); outlcd(D1);
	for(i = 0;i < 3600*WAITSCALE;i++);	
}

void Mode(int com) {
	outlcd(NONE); outlcd(D2 | (com & (D0 | D1)));
}

void Control(int com) {
	outlcd(NONE); outlcd(D3 | (com & (D0 | D1 | D2)));
}

void Cursor(int com) {
	outlcd(D4); outlcd(com & (D3 | D2));
}

void FuncSet() {
	outlcd(D5); outlcd(D3);
}

void PutChar(int c) {
	volatile int	i;
	int	hi, low;

	hi = low = 0;
	if(c & 0x80) hi |= D7;
	if(c & 0x40) hi |= D6;
	if(c & 0x20) hi |= D5;
	if(c & 0x10) hi |= D4;
	if(c & 0x08) low |= D3;
	if(c & 0x04) low |= D2;
	if(c & 0x02) low |= D1;
	if(c & 0x01) low |= D0;
	outlcd(RS | hi);outlcd(RS | low);
	for(i = 0;i < 150*WAITSCALE;i++);	
}

void Locate(int num) {
	int	hi, low;

	hi = low = 0;
	if(num & 0x40) hi |= D6;
	if(num & 0x20) hi |= D5;
	if(num & 0x10) hi |= D4;
	if(num & 0x08) low |= D3;
	if(num & 0x04) low |= D2;
	if(num & 0x02) low |= D1;
	if(num & 0x01) low |= D0;
	outlcd(D7 | hi);outlcd(low);
}

void LcdInit() {
	volatile int	i;

	for(i = 0;i < 100000*WAITSCALE;i++);
	outlcd(D5 | D4);
	for(i = 0;i < 20000*WAITSCALE;i++);
	outlcd(D5 | D4);
	for(i = 0;i < 400*WAITSCALE;i++);	
	outlcd(D5 | D4);
	outlcd(D5);
	FuncSet();
	Control(NONE);
	Clear();
	Mode(NONE);
	Control(LCDDISPLAY | LCDCURSOR);
}

void fix_current() {
	int	i;

	if(current == max_column * 2 + 64) {
		current = max_column + 64;
		Clear();
		memcpy(databuf[0], databuf[1], max_column);
		memcpy(databuf[1], &(databuf[0][max_column]), max_column);
		memcpy(&(databuf[0][max_column]), &(databuf[1][max_column]), max_column);
		for(i = 0;i < max_column;i++) {
			Locate(i);
			PutChar(databuf[0][i]);
			Locate(i + 64);
			PutChar(databuf[1][i]);
		}
		for(i = 0;i < max_column;i++) {
			Locate(i + max_column);
			PutChar(databuf[0][i + max_column]);
			databuf[1][i + max_column] = ' ';
		}
	} else if(current == max_column * 2) {
		current = 64 + max_column;
	} else if(current == max_column + 64) {
		if(max_low == 4) {
			current = max_column;
		} else {
			current = 64;
			Clear();
			memcpy(databuf[0], databuf[1], max_column);
			for(i = 0;i < max_column;i++) {
				Locate(i);
				PutChar(databuf[0][i]);
				databuf[1][i] = ' ';
			}
		}
	} else if(current == max_column) {
		current = 64;
	}
}

int lcd_putch(char code) {
	int		i, n;
	unsigned char	c;

	c = code;
	if(c == 0x7e) c = 0xf3;
	if(c == 0x8e) return -1;
	if(c == 0x0d) return -1;
	if(c == 8) {
		if(current == 0) {
		} else if(current == 64) {
			current = max_column - 1;
		} else if(current == max_column) {
			current = max_column - 1 + 64;
		} else if(current == (max_column + 64)) {
			current = max_column * 2 - 1;
		} else {
			current--;
		}
		Locate(current);
	} else if(c == '\n') {
		if(current < max_column) current = max_column;
		else if(current < max_column * 2) current = max_column * 2;
		else if(current < 64) current = 0;
		else if(current < 64 + max_column) current = max_column + 64;
		else if(current < max_column * 2 + 64) current = max_column * 2 + 64;
		else current = 0;
		fix_current();
		Locate(current);
	} else if(c >= 0x20) {
		Locate(current);
		PutChar(c);
		if(current < 64)
			databuf[0][current] = c;
		else
			databuf[1][current - 64] = c;
		current++;
		fix_current();
	}
	return 0;
}

int lcd_write(char *buffer, int size) {
	int	i;

	for(i = 0;i < size;i++) lcd_putch(buffer[i]);
}

void lcd_clear() {
	current = 0;
	Clear();
}

void lcdsetup(int num, volatile unsigned REGTYPE *regptr, unsigned REGTYPE *mask) {
	int	i, n, low, column;

	low = (num >> 8) & 0xff;
	column = num & 0xff;
	if(low == 4) max_low = low;
	else max_low = 2;
	max_column = column;
	SerialMode = 0;
	if(mask[0] == 0xff) {
		for(i = 0;i < 3;i++) masks164[i] = mask[i + 1];
		for(i = 0;i < 7;i++) masks[i] = mask[i + 4];
	} else {
		for(i = 0;i < 7;i++) masks[i] = mask[i];
	}
	lcdreg = regptr;
	*lcdreg = 0;
	LcdInit();
	current = 0;
	n = (max_low == 4) ? max_column * 2 : max_column;
	for(i = 0;i < n;i++) {
		databuf[0][i] = ' '; 
		databuf[1][i] = ' '; 
	}
}
