//#include <stdlib.h>
#define ASCII	0
#define BINARY	1
#define SIO	2
#define LCD	4
#define PS2	4
#define MEM	6

#define	CS0	0x01
#define	CS1	0x02
#define	CS2	0x04
#define	CS3	0x08
#define	CS4	0x10
#define	CS5	0x20
#define	CS6	0x40
#define	CS7	0x80

static	unsigned char	binary = 0;
static	unsigned char	outdev = SIO;
static	unsigned char	indev = SIO;
static	char	*membuf;
const unsigned char hextbl[] = "0123456789ABCDEF";
int	memoffset, memmax;
int sio_write(char*, int);
int sio_read(char*, int);
int key_read(char*, int);
int lcd_write(char*, int);
int ne2000_write(char*, int, int);
int ne2000_read(char*, int, int);
int ne2000_flush(int);
int ne2000_input(int);
int ne2000_get();
int ne2000_setmac(char*);
int ne2000_getmac(char*);
int ne_setup();
int cs_setup(int);
extern  volatile unsigned char	*ne_base_addr;
extern  int	neregmul;
extern	char	ether_enable;
extern	char	mac_address[];

int atohex(char *str) {
	int	val, i, mi;
	char	c;
	
	val = i = 0;
	while(str[i] <= ' ') i++;
	mi = i;
	while(str[i] > ' ' && (i - mi) < 8) {
		val <<= 4;
		c = str[i];
		if(c >= '0' && c <= '9') {
			val += c - '0';
		} else if(c >= 'a' && c <= 'f') {
			val += c - 'a' + 10;
		} else if(c >= 'A' && c <= 'F') {
			val += c - 'A' + 10;
		} else {
			return -1;
		}
		i++;
	}
	return val;
}

int atodec(unsigned char *str) {
	int		val, i, mi;
	unsigned char	c;

	val = i = 0;
	while(str[i] <= ' ') i++;
	mi = i;
	if(str[i] == '-') i++;
	while(str[i] > ' ' && (i - mi) < 12) {
		val *= 10;
		c = str[i];
		if(c >= '0' && c <= '9') {
			val += c - '0';
		} else {
			return -1;
		}
		i++;
	}
	if(str[mi] == '-') val = val * (-1);
	return val;
}

int write_mode(int flag, char *buffer, int max) {
	switch(flag) {
	case ASCII:
	case BINARY:
		binary = flag;
		break;
	case MEM:
		membuf = buffer;
		memoffset = 0;
		memmax = max; 
	case SIO:
	case LCD:
		outdev = flag;
		break;
	default:
		return -1;
	}
	return 0;
}

int read_mode(int flag) {
	switch(flag) {
	case SIO:
	case PS2:
		indev = flag;
		break;
	default:
		return -1;
	}
	return 0;
}

int write_device(char *str, int size) {
	switch(outdev) {
	case SIO:
		sio_write(str, size);
		break;
	case LCD:
		lcd_write(str, size);
		break;
	case MEM:
		if(memoffset + size > memmax-3 ) return -1;
		memcpy(&(membuf[memoffset]), str, size);
		memoffset += size;
		membuf[memoffset+1] = 0 ; // by kishida koji (2009.11.1)  
		break;
	default:
		return -1;
	}
	return 0;
}	

int sys_write(char *str, int size) {
	if(binary == BINARY) hex_write(str, size);
	else write_device(str, size);
	return 0;
}

int write_string(char *str) {
	int	n;

	n = strlen(str);
	return sys_write(str, n);
}

int write_decimal(int value) {
	int	a, i, f, dec;
	char	c;

	dec = value;
	c = '-';
	f = 0;
	if(dec < 0) {
		write_device(&c, 1);
		dec = ~dec + 1;
	}
	for(i = 100000000;i >= 100;i /= 10) {
		a = (dec % i) / (i / 10);
		if(a > 0) f = 1;
		if(f == 1) write_device((char*)(&hextbl[a]), 1);
	}
	a = dec % 10;
	write_device((char*)(&hextbl[a]), 1);
	return 0;
}

int sys_read(char *str, int size) {
	int	n;

	if(indev == PS2) {
		n = key_read(str, size);
	}
	else n = sio_read(str, size);
	return n;
}

void flush_line(char *data, int index) {
	unsigned int	i, n;
	unsigned char	c;

	n = strlen(data);
	c = 8;
	for(i = index;i < n;i++) sys_write(&(data[i]), 1);
	sys_write(" ", 1);
	for(i = index;i < n;i++) sys_write(&c, 1);
}

void insert_char(char *data, int index, char c) {
	int	i, n;

	n = strlen(data);
	for(i = n;i >= index;i--) data[i + 1] = data[i];
	data[index] = c;
}

void delete_char(char *data, int index) {
	int	i, n;

	n = strlen(data);
	for(i = index;i <= n;i++) data[i - 1] = data[i];
}

int read_string(char *data, int size) {
	unsigned int	n, index;
	unsigned char	c;

	index = 0;
	data[index] = 0;
	while(1) {
		n = sys_read(&c, 1);
		if(n < 1) continue;
		if(c == '\r') {
			break;
		}
		if(c == 2 && index > 0) {
			c = 8;
			sys_write(&c, 1);
			flush_line(data, --index);
			sys_write(&c, 1);
		} else if(c == 6 && data[index] != 0) {
			flush_line(data, index++);
			sys_write(&c, 1);
		} else if(c == 8 && index > 0) {
			delete_char(data, index);
			c = 8;
			sys_write(&c, 1);
			flush_line(data, --index);
			sys_write(&c, 1);
		} else if(c == 0x7f) {
			delete_char(data, index + 1);
			flush_line(data, index);
			c = 8;
			sys_write(&c, 1);
		} else if(c >= 0x20 && ((index < size) || size == -1)) {
			insert_char(data, index, c);
			flush_line(data, index++);
		}
	}
	n = strlen(data);
	return n;
}

int hex_write(char *str, int size) {
	unsigned int	i, val;
	unsigned char	buf[2];

	for(i = 0;i < size;i++) {
		val = (int)(str[i]) & 0xff;
		buf[0] = hextbl[(val >> 4) & 0x0f];
		buf[1] = hextbl[val & 0x0f];
		write_device(buf, 2);
	}
}

void nop() {
}

#if defined (H8_300H) || defined(SH2)
int ether_write(char *data, int index, int size) {
	return ne2000_write(data, index, size);
}

int ether_read(char *data, int index, int size) {
	return ne2000_read(data, index, size);
}

int ether_flush(int sendsize) {	
	return ne2000_flush(sendsize);
}	

int ether_get() {
	return ne2000_get();
}

void ether_init(int addr) {
	if( addr == 0 ){	// kishida koji (2009/12/6)
		ip_init() ; 	// kishida koji (2009/12/6)
		ne_init() ; 	// kishida koji (2009/12/6)
	}
	else{		// kishida koji (2009/12/6)
	ne_base_addr = (volatile unsigned char*)addr;
	if(ne_base_addr[0] == ne_base_addr[1]) neregmul = 2;
	else neregmul = 1;
	ne_setup();
	};		// kishida koji (2009/12/6)
}

void ether_setup() {
	ne2000_init();
}

int get_ether_status() {
	return ether_enable;
}

char* get_mac_address() {
	return mac_address;
}

int ether_setmac(char *mac) {
	return ne2000_setmac(mac);
}

int ether_getmac(char *mac) {
	return ne2000_getmac(mac);
}
#endif
#ifdef H8_3664
void sleep(int time) {
	volatile int	i, w;
	
	for(i = 0;i < time;i++) for(w = 0;w < 100000;w++);	
}
#endif
