#include <h8/syscall.h>
#include <h8/printf.h>
#include <string.h>
#define MAX_BUF	88
#define MAX_ARG	40
#define NONE	0
#define COMMAND	1
#define SYNTAX	2
#define IOERR	3
#define DSKFULL	4
#define CMDFULL	5
#define NOFILE	6
#define EEPCHK	1
#define FSCHK	2
int		fd, eepsize;
char		fsready;
static char	prompt[16];
extern int	user_start;

typedef struct {
	int	(*function)(int , char**);
	char	command[9];
	char	min, max, opt;
} function_table;

void init() {
	sio_speed(57600, FREQ);
	fd = 0;
	eepsize = get_eepsize();
	fsready = (easy_max() == 0) ? 0 : 1;
	strcpy(prompt, "\nMSFOS >");
	printf("Micro System, File Operating System. Ver 1.0", eepsize);
	if(eepsize > 0) printf("EEPROM size %d[byte]", eepsize);
	putchar('\n');
}	

int read_line(char *buffer, char **argv) {
	int		i, argc;
	unsigned char	*ptr;

	ptr = buffer;
	for(;;) {
		if((int)ptr >= (int)(buffer + MAX_BUF)) ptr = &buffer[MAX_BUF - 1];
		*ptr = getchar();
		if(*ptr >= 0x80) continue;
		if(*ptr == '\r') break;
		if(*ptr == 0x7f) *ptr = 8;
		if(*ptr == 8) {
			if((int)ptr > (int)buffer) {
				putchar(*ptr--);
			}
		} else {
			putchar(*ptr++);
		}
	}
	*ptr = 0;
	if(buffer[0] == 0) return 0;
	ptr = buffer;
	i = argc = 0;
	argv[argc++] = &(buffer[i]);
	while(buffer[i] != 0) {
		if(buffer[i] == ' ') {
			buffer[i] = 0;
			argv[argc++] = &(buffer[i + 1]);
		}
		i++;
	}
	return argc;
}

int err_message(int id) {
	static const char message[7][20] = {
		"Command not found.",
		"Syntax error.",
		"Disk I/O error.",
		"Disk is full.",
		"Buffer is full.",
		"File not found.",
		"Unknown error."
	};
	int	i;

	i = (id > 6) ? 6 : id - 1;
	printf("ERROR[%s]\n", message[i]);
}

int dump(int argc, char **argv) {
	int	i, n, num;
	char	*p;

	num = atohex(argv[1]);
	n = num + 16 * 8;
	while(num < n) {
		p = (char*)num;
		printf("%08x : ", num);
		for(i = 0;i < 16;i++) {
			if(i == 8) printf("  ");
			printf("%02x", (int)(*p++) & 0xff);
		}
		putchar('\n');
		num += 16;
	}
	return NONE;
}

int disk_dump(int argc, char **argv) {
	int	i, n, num;
	char	c;

	num = atohex(argv[1]);
	n = num + 16 * 8;
	while(num < n) {
		printf("%08x : ", num);
		for(i = 0;i < 16;i++) {
			eep_read(&c, num + i, 1);
			if(i == 8) printf("  ");
			printf("%02x", (int)c & 0xff);
		}
		putchar('\n');
		num += 16;
	}
	return NONE;
}

int disk_write(int argc, char **argv) {
	int	i, n, addr, num;
	char	buf[32];

	if(fd < 1) {
		putchar('.');
		return NONE;
	}
	if(argc == 2) return SYNTAX;
	addr = atohex(argv[1]);
	num = atohex(argv[2]);
	if(addr == -1 || num == -1 || argc != num + 3 || num > 32) return SYNTAX;
	for(i = 0;i < num;i++) buf[i] = (char)atohex(argv[i + 3]);
	n = (addr + num - 1) >> 6;
	n <<= 6;
	if(n > addr) {
		easy_write(fd, buf, addr, n - addr);
		easy_write(fd, buf, n, num + addr - n);
	} else {
		easy_write(fd, buf, addr, num);
	}
	putchar('.');
	return NONE;
}

int exec(int argc, char **argv) {
	int	ret;

	ret = easy_exec(argc, argv, (char*)(&user_start));
	printf(" executed.\n");
	return (ret == 0) ? NONE : COMMAND;
}

int dir(int argc, char **argv) {
	int	i, n, ret;
	char	buf[16];

	n = easy_max();
	for(i = 0;i < n;i++) {
		bzero(buf, 16);
		ret = easy_dir(i, buf);
		if(ret >= 0) {
			printf("%-14s%d[byte]\n", buf, ret);
		}
	}
}

int disk_format(int argc, char **argv) {
	easy_format('.');
	fsready = (easy_max() == 0) ? 0 : 1;
	return NONE;
}

int newfile(int argc, char **argv) {
	int	num, ret;

	fd = -1;
	num = atohex(argv[2]);
	ret = easy_create(argv[1], num);
	if(ret == -1) return DSKFULL;
	printf("file created.\n");
	fd = ret;
	return NONE;
}

int del_file(int argc, char **argv) {
	int	ret;

	ret = easy_del(argv[1]);
	if(ret == -1) return NOFILE;
	printf("file deleted.\n");
	return NONE;
}

int find_file(int argc, char **argv) {
	int	ret;

	if(argc == 1) {
		fd = 0;
	} else if(argc == 2) {
		fd = -1;
		ret = easy_find(argv[1]);
		if(ret == -1) return NOFILE;
		fd = ret;
	}
	return NONE;
}

int edit_byte(int argc, char **argv) {
	int	addr, num;
	char	c;

	addr = atohex(argv[1]);
	num = atohex(argv[2]);
	if(addr == -1 || num == -1 || addr >= eepsize) return SYNTAX;
	eep_read(&c, addr, 1);
	c = num & 0xff;
	eep_write(addr, &c, 1);
	printf(" updated.\n");
	return NONE;
}

int set_byte(int argc, char **argv) {
	int	num;
	char	c, *addr;

	addr = (char*)atohex(argv[1]);
	num = atohex(argv[2]);
	if(addr == (char*)-1 || num == -1) return SYNTAX;
	addr[0] = num;
	printf(" updated.\n");
	return NONE;
}

int clear_memory(int argc, char **argv) {
	int	num;
	char	*addr;

	addr = (char*)atohex(argv[1]);
	num = atohex(argv[2]);
	if(addr == (char*)-1 || num == -1)  return SYNTAX;
	bzero(addr, num);
	printf(" cleared.\n");
}

int main() {
	static const function_table table[] = {
		{dump,		"dump",	    2, 2, NONE},
		{disk_dump,	"diskdump", 2, 2, EEPCHK},
		{edit_byte,	"editbyte", 3, 3, EEPCHK},
		{set_byte,	"setbyte",  3, 3, NONE},
		{clear_memory,	"zeromem",  3, 3, NONE},
		{disk_write,	"eepdata",  1, MAX_ARG, FSCHK},
		{dir, 		"dir",      1, 1, FSCHK},
		{del_file,	"del",      2, 2, FSCHK},
		{newfile,	"newfile",  3, 3, FSCHK},
		{find_file,	"file",     1, 2, FSCHK},
		{disk_format,	"format",   1, 1, NONE},
		{exec, 		"",         1, MAX_ARG, NONE}
	};
	static char	buffer[MAX_BUF], *argv[MAX_ARG];
	int		ret, i, argc;

	init();
	for(;;) {
		printf(prompt);
		argc = read_line(buffer, argv);
		putchar('\n');
		if(argc >= MAX_ARG) {
			err_message(CMDFULL);
			continue;
		}
		for(i = 0;;i++) {
			if(strcmp(argv[0], table[i].command) == 0 || table[i].command[0] == 0) {
				if(argc >= table[i].min && argc <= table[i].max &&
				   ((table[i].opt == FSCHK && fsready > 0) ||
				    (table[i].opt == EEPCHK && eepsize > 0) ||
				    (table[i].opt == NONE))) {
					ret = (*(table[i].function))(argc, argv);
					if(ret > NONE) err_message(ret);
				} else {
					err_message(SYNTAX);
				}
				break;
			}
			if(table[i].command[0] == 0) break;
		}
	}
}
