#include "regdef.h"
#ifdef H8_3067
  #define SIONUM 3
  #define STDSIO 1
  #define REGINC 8
#endif
#ifdef H8_3048
  #define SIONUM 2
  #define STDSIO 1
  #define REGINC 8
#endif
#ifdef SH2
  #define SIONUM 2
  #define STDSIO 1
  #define REGINC 16
#endif
#ifdef H8_3664
  #define SIONUM 1
  #define STDSIO 0
  #define REGINC 8
#endif

#define TDRE	0x80
#define RDRF	0x40
#define ORER	0x20
#define FER	0x10
#define PER	0x08
#define TEND	0x04
#define TIE	0x80
#define RIE	0x40
#define TE	0x20
#define RE	0x10

#define SMR	0
#define BRR	1
#define SCR	2
#define TDR	3
#define SSR	4
#define RDR	5

volatile unsigned char *scr_base;

void sio_init() {
	int	i, w;

	scr_base = &SMR0;
	for(i = 0;i < SIONUM;i++,scr_base += REGINC) {
		scr_base[SMR] = 0;
		scr_base[SCR] = 0;
#ifdef SH2
		scr_base[BRR] = 14;
#endif
#ifdef H8_3067
		scr_base[BRR] = 10;
#endif
#ifdef H8_3048
		scr_base[BRR] = 8;
#endif
#ifdef H8_3664
		PMR1 = 0x0e;
		scr_base[BRR] = 8;
	#endif
		for(w = 0;w < 280;w++);
		scr_base[SCR] = RE | TE;
		w = scr_base[SSR];
		scr_base[SSR] = TDRE;
	}
	scr_base = &SMR0;
	scr_base = &(scr_base[STDSIO * REGINC]);
}

int sio_write(char *buffer, int size) {
	int	i;

	for(i = 0;i < size;i++) {
		if(buffer[i] == '\n') {
			while(!(scr_base[SSR] & TDRE));
			scr_base[TDR] = '\r';
#ifndef H8_3664
			scr_base[SSR] &= ~TDRE;
#endif
		}
		while(!(scr_base[SSR] & TDRE));
		scr_base[TDR] = buffer[i];
#ifndef H8_3664
		scr_base[SSR] &= ~TDRE;
#endif
	}
	while(!(scr_base[SSR] & TEND));
	return 0;
}

int sio_read(char *buffer, int size) {
	int	i, n;
	char	c;

	n = 0;
	for(i = 0;i < size;i++) {
		if(scr_base[SSR] & (ORER | FER | PER)) {
			scr_base[SSR] &= ~(ORER | FER | PER);
			return 0;
		}
		if(scr_base[SSR] & RDRF) {
			buffer[n++] = scr_base[RDR];
#ifndef H8_3664
			scr_base[SSR] &= ~RDRF;
#endif
		} else {
			return n;
		}
	}
	return n;
}

int sio_speed(int speed, int freq) {
#ifdef SH2
	unsigned int	i, x, y, b, s;
#else
	int	i, x, y, b, s;
#endif

	s = 32;
	for(i = 0;i < 4;i++) {
		x = freq * 10000000;
		y = s * speed;
		b = x / y + 5;
		b = b / 10 - 1;
		if(b <= 255) break;
		s <<= 2;
	}
	if(i == 4) return -1;
	scr_base[SMR] &= ~0x03;
	scr_base[SMR] |= i;
	scr_base[BRR] = b;
	return 0;
}

int sio_change(unsigned int sci_number) {
	if(sci_number >= SIONUM) return -1;
	scr_base = &SMR0;
	scr_base = &(scr_base[sci_number * REGINC]);
	return 0;
}

