#include <h8/reg3067.h>
#define TDRE	0x80
#define RDRF	0x40
#define ORER	0x20
#define FER	0x10
#define PER	0x08
#define TEND	0x04
#define TIE	0x80
#define RIE	0x40
#define TE	0x20
#define RE	0x10
#define TxD0	0x01
#define TxD1	0x02
#define RxD0	0x04
#define RxD1	0x08

#define MAX_BUF 64
static char sci_buffer[MAX_BUF];
static int  index;
static int  i, n, d;

void sio_init() {
	P9DDR = 0;
	SCR1 = RIE;
	SMR1 = 0;
/*	BRR1 = 64;	 20MHz - 9600bps */
/*	BRR1 = 32;	 20MHz - 19200bps */
/*	BRR1 = 15;	 20MHz - 38400bps */
/*	BRR1 = 10;	 20MHz - 57600bps */
	BRR1 = 10;
	for(i = 0;i < 280;i++);
	SCR1 |= RE | TE;

	index = 0;
}

int sio_write(char *buffer, int size) {
/*	P9DDR &= ~TxD1;
	SCR1 |= TE;*/
	i = SSR1;
	SSR1 = TDRE;
	for(i = 0;i < size;i++) {
		if(buffer[i] == '\n') {
			while(!(SSR1 & TDRE));
			TDR1 = '\r';
			SSR1 &= ~TDRE;
		}
		while(!(SSR1 & TDRE));
		TDR1 = buffer[i];
		SSR1 &= ~TDRE;
	}
	while(!(SSR1 & TEND));
/*	P9DDR |= TxD1;
	P9DR &= ~TxD1;
	SCR1 &= ~TE;*/
	return 0;
}

int sio_read(char *buffer, int size) {
	for(i = 0;i < size;i++) {
		if(i >= index) break;
		buffer[i] = sci_buffer[i];
	}
	n = i;
	d = 0;
	for(;i < index;i++) {
		sci_buffer[d++] = sci_buffer[i];
	}
	index = d;
	if(index < MAX_BUF) {
		SCR1 |= RE;
	}
	return n;
}

int sio_speed(int speed) {
	switch(speed) {
	case 1200:
		SMR1 &= ~0x03;
		SMR1 |= 0x01;
		BRR1 = 129;
		break;
	case 2400:
		SMR1 &= ~0x03;
		SMR1 |= 0x01;
		BRR1 = 64;
		break;
	case 4800:
		SMR1 &= ~0x03;
		BRR1 = 129;
		break;
	case 9600:
		SMR1 &= ~0x03;
		BRR1 = 64;
		break;
	case 19200:
		SMR1 &= ~0x03;
		BRR1 = 32;
		break;
	case 31250:
		SMR1 &= ~0x03;
		BRR1 = 19;
		break;
	case 38400:
		SMR1 &= ~0x03;
		BRR1 = 15;
		break;
	case 57600:
		SMR1 &= ~0x03;
		BRR1 = 10;
		break;
	default:
		return -1;
	}
	return 0;
}

#pragma interrupt
int	int_eri1() {
	SSR1 &= ~(ORER | FER | PER);
}

#pragma interrupt
int	int_rxi1() {
	sci_buffer[index++] = RDR1;
	if(index >= MAX_BUF) {
		SCR1 &= ~RE;
	}
	SSR1 &= ~RDRF;
}
