#include "regdef.h"
#ifdef H8_3067
 static void (*timer1_proc)();
 static void (*timer2_proc)();
 static void (*timer3_proc)();
#endif

#if defined (H8_300H) || defined(SH2)
 volatile int  time_over_arp;
 volatile int  sequence;
 void ip_interval();
 extern char ether_enable;
#endif

static volatile int down_count;
int int_regist(int, void (*)());

int timer1_regist(void (*proc)(), unsigned int counter) {
#ifdef H8_3067
	int	data;

	if(timer1_proc != (void*)0) return -1;
	data = TCSR1_8; TCSR1_8 = 0;
	TCNT1_8 = 0;
	TCORA1_8 = counter;
	TCR1_8 &= ~0x1f;
	TCR1_8 |= 0x4b;
	timer1_proc = proc;
	return 0;
#else
	return -1;
#endif
}

int timer1_unregist() {
#ifdef H8_3067
	TCR1_8 &= ~0x40;
	timer1_proc = (void*)0;
#endif
}

int timer2_regist(void (*proc)(), unsigned int counter) {
#ifdef H8_3067
	int	data;

	if(timer2_proc != (void*)0) return -1;
	data = TCSR2_8; TCSR2_8 = 0;
	TCNT2_8 = 0;
	TCORA2_8 = counter;
	TCR2_8 &= ~0x1f;
	TCR2_8 |= 0x4b;
	timer2_proc = proc;
	return 0;
#else
	return -1;
#endif
}

int timer2_unregist() {
#ifdef H8_3067
	TCR2_8 &= ~0x40;
	timer2_proc = (void*)0;
#endif
}

int timer3_regist(void (*proc)(), unsigned int counter) {
#ifdef H8_3067
	int	data;

	if(timer3_proc != (void*)0) return -1;
	data = TCSR3_8; TCSR3_8 = 0;
	TCNT3_8 = 0;
	TCORA3_8 = counter;
	TCR3_8 &= ~0x1f;
	TCR3_8 |= 0x4b;
	timer3_proc = proc;
	return 0;
#else
	return -1;
#endif
}

int timer3_unregist() {
#ifdef H8_3067
	TCR3_8 &= ~0x40;
	timer3_proc = (void*)0;
#endif
}

int sleep(int ms) {
#ifdef H8_3067
	down_count = ms * 20;
#endif
#ifdef H8_3048
	down_count = ms * 24;
#endif
#ifdef SH2_7045
	down_count = ms * 21;
#endif
#ifdef SH2_7050
	down_count = ms * 15;
#endif
#ifdef H8_3664
	down_count = ms * 49;
#endif
	while(down_count > 0);
}

#if defined (H8_3048) || defined(SH2)
void wovi_proc() {
	static char	w;

	TCSR_WR = 0xa500 | 0x1c;
	w = TCSR_RD;
	TCSR_WR = 0xa500 | 0x3c;
	sequence++;
	if(down_count > 0) down_count--;
	if(time_over_arp > 0) time_over_arp--;
	if(ether_enable != 0) ip_interval();
}
#endif

#ifdef H8_3664
void timerv_proc() {
	if(down_count > 0) down_count--;
	TCSRV &= ~0x20;
}
#endif

#ifdef H8_3067
void cmia0_proc() {
//	if(TCSR0_8 & 0x40) {   // kishida
		sequence++;
		if(down_count > 0) down_count--;
		if(time_over_arp > 0) time_over_arp--;
		if(ether_enable != 0) ip_interval();
		TCSR0_8 &= ~0x40;
//	} ;  // kishida
//	TCNT = 0x5a00;  // kishida 2008/9/17
	return ; // kishida 2008/9/24
}

void cmiab1_proc() {
	static int count = 0;
	
	if(TCSR1_8 & 0x40) {
		if(timer1_proc != (void*)0) (*timer1_proc)();
		TCSR1_8 &= ~0x40;
	}
}

void cmia2_proc() {
	static int count = 0;
	
	if(TCSR2_8 & 0x40) {
		if(timer2_proc != (void*)0) (*timer2_proc)();
		TCSR2_8 &= ~0x40;
	}
}

void cmiab3_proc() {
	static int count = 0;
	
	if(TCSR3_8 & 0x40) {
		if(timer3_proc != (void*)0) (*timer3_proc)();
		TCSR3_8 &= ~0x40;
	}
}
#endif

int timer_init() {
#ifndef H8_3664
	sequence = 1;
#endif
#ifdef H8_3067
	int	data;

	int_regist(36, cmia0_proc);
	int_regist(38, cmiab1_proc);
	int_regist(40, cmia2_proc);
	int_regist(42, cmiab3_proc);
	timer1_proc = (void*)0;
	timer2_proc = (void*)0;
	timer3_proc = (void*)0;
	data = TCSR0_8; TCSR0_8 = 0;
	TCNT0_8 = 0;
	TCORA0_8 = 12;
	TCR0_8 = 0x4b;
#endif
#ifdef H8_3048
	int_regist(20, wovi_proc);
#endif
#ifdef SH2_7045
	int_regist(152, wovi_proc);
	IPRH = 0xf000;
#endif
#ifdef SH2_7050
	int_regist(164, wovi_proc);
	IPRH = 0x000f;
#endif
#if defined (H8_3048) || defined(SH2)
	TCSR_WR = 0xa500 | 0x1c;
	TCNT_WR = 0x5a00;
	TCSR_WR = 0xa500 | 0x3c;
#endif
#ifdef H8_3664
	TCRV0 = 0x23;		/* Сեx128 */
	TCRV1 = 0x01;		/* å x128 */
	TCNTV = 0;
	int_regist(22, timerv_proc);
#endif
}
