#include "tcpip.h"
int	ether_write(char*, int, int);
int	ether_flush(int);
int	ether_read(char*, int, int);
int	get_ether_status();
ip_header* set_ip_header(char, int, int, int, char);
char*	query_mac(int);
#ifdef LARGE
  #define MAX_UDPPROC	60
#else
  #define MAX_UDPPROC	20
#endif
int	(*udpproc[MAX_UDPPROC])(unsigned short, struct sockaddr_in*);
static	int	(*saveproc[MAX_UDPPROC])(unsigned short, struct sockaddr_in*);
static	udp_data_offset, udp_data_size;
static  struct sockaddr_in	destaddr;
static  unsigned short		destport;
extern struct mac_info	mytable;
extern unsigned short	tmp_port;
extern char*		query_mac(int);
extern unsigned char	sock_flag[];
extern char		ip_enable;

int set_udp_header(unsigned short src, unsigned short bind, int header_size, int data_size) {
	udp_header	udp;

	if(bind == 0) udp.src_port = tmp_port++;
	else udp.src_port = bind;
	udp.dest_port = src;
	udp.data_size = sizeof(udp_header) + data_size;
	udp.checksum = 0;
	ether_write((char*)&udp, sizeof(ether_header) + header_size, sizeof(udp_header));
}

int udp(ip_header *ip) {
	unsigned int	src_ip, *p;
	unsigned int	header_size, data_size;
	udp_header	udp;
	struct sockaddr_in addr;
	int	i, ret;

	header_size = ((int)ip->version_size & 0xf) * 4;
	ether_read((char*)&udp, sizeof(ether_header) + header_size, sizeof(udp_header));
	p = (int*)ip->src_ip; src_ip = *p;
	data_size = udp.data_size - sizeof(udp_header);
	for(i = 0;i < MAX_UDPPROC;i++) {
		if(udpproc[i] != 0) {
			addr.sin_port = udp.src_port;
			addr.sin_addr = src_ip;
			destaddr = addr;
			destport = udp.dest_port;
			udp_data_offset = sizeof(ether_header) + header_size + sizeof(udp_header);
			udp_data_size = data_size;
			ret = (*udpproc[i])(udp.dest_port, &addr);
			udp_data_offset = 0;
			if(ret == 0) return 0;
		}
	}
	return -1;
}

int udp_read(char *data, int len) {
	int size;
	
	if(udp_data_offset == 0) return -1;
	if(udp_data_size > len) size = len;
	else size = udp_data_size;
	ether_read(data, udp_data_offset, size);
	return size;
}

void set_udpaddr(struct sockaddr_in *addr, unsigned short port) {
	destaddr = *addr;
	destport = port;
}

int udp_write(char *data, int len) {
	char		*mac, sin_port;
	char		broadcast[] = {0xff,0xff,0xff,0xff,0xff,0xff};

	if(ip_enable == 0) return -1;
        if(destaddr.sin_addr == 0xFFFFFFFF){
                mac = broadcast;
        }else{
        	if((mac = query_mac(destaddr.sin_addr)) == 0) return -1;
        }
	set_ether_header(mac, mytable.mac, IP_TYPE);
	set_ip_header(IP_UDP, destaddr.sin_addr, 20, sizeof(udp_header) + len, 0);
	set_udp_header(destaddr.sin_port, destport, 20, len);
	ether_write(data, sizeof(ether_header) + 20 + sizeof(udp_header), len);
	ether_flush(sizeof(ether_header) + 20 + sizeof(udp_header) + len);
	return len;
}

int udp_regport(int (*proc)(unsigned short, struct sockaddr_in*)) {
	int	i;
	
	for(i = 0;i < MAX_UDPPROC;i++) {
		if(udpproc[i] == 0) {
			udpproc[i] = proc;
			return i;
		}
	}
	return -1;
}

int udp_unregport(int (*proc)(unsigned short, struct sockaddr_in*)) {
	int	i;
	
	for(i = 0;i < MAX_UDPPROC;i++) {
		if(udpproc[i] == proc) {
			udpproc[i] = 0;
			return 0;
		}
	}
	return -1;
}

void save_regport() {
/* 2003/03/07 change start */
	memcpy(&saveproc[0], &udpproc[0], sizeof(udpproc));
/* 2003/03/07 change end */
}

void recover_regport() {
/* 2003/03/07 change start */
	memcpy(&udpproc[0], &saveproc[0], sizeof(udpproc));
/* 2003/03/07 change end */
}

int udp_init() {
/* 2003/03/07 change start */
	memset(&udpproc[0], 0, sizeof(udpproc));
/* 2003/03/07 change end */
	udp_data_offset = 0;
}
